/*
   BILLGPS.c - main module for VxD BILLGPS

   Free software Copyright (c) 1996 Lew Perin

   This VxD acts as "server" for the billgp "protocol" allowing the "client"
   billgPC to reach around MS/TCP to transact a bootp query/response.

   The basic purpose of this driver is to trap an incoming bootp
   reply and pass it to the caller app.  This is necessary because
   billgPC, which uses WinSock to transmit its bootp request, would
   never see the incoming reply if it relied on WinSock for input in
   a situation where the client and server disagreed on the client's
   IP address *and* the server didn't comply with the broadcast flag
   handling called out in RFC 1542.

   The original design of this driver was indebted to Russinovich and
   Cogswell's VCMON.  In that version we hooked NdisIndicateReceive
   rather than presenting ourselves as an NDIS transport.
   When we realized that not every netcard driver calls
   NdisIndicateReceive (some take the shortcut of vectoring to the
   NDIS binding context's receive handler) we did a radical redesign;
   now it's more of an orthodox NDIS transport.  The armature for the
   redesigned driver was Thomas F. Divine's transport driver skeleton
   TRANS31, which is distributed with VtoolsD.  The design for this
   driver owes much to conversations with Thomas F. Divine and P. Singh,
   as well as code Singh was nice enough to show me.

   Later, though, the service hooking tactic returned as a supplement.
   We now hook NdisSend so we can snag the outgoing bootp request and
   zero the IP source address and adjust the checksums correspondingly.
   The reason for this is to use the bootp exemption for IP stacks'
   "Martian address filtering" as discussed in RFC 1542's section 4.1.
   This became necessary when it transpired that if billgPC and a
   Novell file server's IP stack disagreed on the client's subnet the
   bootpd on the file server would never see the billgPC request.
   Note that all DOS/Windows 3.x bootp clients zero the IP source address
   in a request, so this shouldn't cause any trouble.

   As of 1.23, we're past the main disadvantage of NDIS protocolhood:
   we're now dynamically loaded rather than being loaded by NDIS when
   Windows starts up.
*/

/*
   Revision History:

   Version  Date      Reason
   -------  --------  ---------------------------------------------------
   1.02       9/1/96  Hooking NdisIndicateReceive, we can snag bootp
                      replies on *almost* all netcards.
   1.03      9/23/96  First stable version as NDIS transport.
   1.05      9/26/96  Implement new capabilities: promiscuous mode for
                      diagnostic purposes, get driver version number,
                      get MAC address.
   1.07      9/28/96  New diagnostic capability to dump the VxD's context.
                      Became more careful about avoiding "NDIS time bombs"
                      where we give NDIS an address on the stack where it
                      may put a *delayed* output datum.
   1.14     11/12/96  New IOCTL to return the NDIS medium type translated
                      into a bootp hardware type.  Started using
                      NdisCompleteBindAdapter on asynchronous opens.
                      Started accepting frame-level broadcasts (as opposed
                      to IP-level broadcasts, which were always OK.)
   1.15     11/19/96  Fixed switch without cases that ruined our medium
                      identification.  Fixed (thanks to Dan Lanciani) our
                      assumption that NdisOpen's supplemental error status
                      meant something when the primary status was OK.
   1.16      12/3/96  Realized that in token ring the first byte of the
                      MAC address's most significant bit may change.
   1.17      12/6/96  Fixed our handling of IEE LLC/SNAP that's typical
                      in token ring IP encapsulation.
   1.20     12/10/96  Made our promiscuous mode a more ambitious packet
                      sniffer by allowing all packets withing a size range
                      rather than just those that are for us.
   1.21      1/25/97  Partially reintroduced our old NDIS service hooking
                      tactic: this time we're hooking NdisSend() so we can
                      replace the WinSock-installed source IP address in
                      the outgoing bootp request with the "Martian" address
                      0.0.0.0.
   1.22      2/16/97  So we won't freeze OSR2 we now make sure when our 
                      ProtocolOpenAdapterComplete is called by NDIS it calls
                      NdisCompleteBindAdapter even if our ProtocolBindAdapter
                      never sees NDIS_STATUS_PENDING.
   1.23      6/13/97  We're now dynamically loaded using the Ingle/Chlap
                      technique that requires a "bind" call after connecting
                      to the VxD.
   1.24      6/30/97  Taking on more of the responsibilities of a normal
                      driver, we now read packets directly into the app.
   1.25      7/26/97  To harmonize with the coming NT driver, we now find
                      SHARED.H in the app directory and devine IOCTL codes
                      in a way that should pass muster under NT.
   1.26     10/27/97  Fixed bug in unbind handler that caused an endless
                      loop if there was an outstanding non-promiscuous read.
*/

/*
  A note on variable names: It won't take much time reading this code
  to see that we're utterly uninterested in the Microsoft variable
  naming conventions that are so widely seen in 95 and NT driver code.
  We find it usually unnecessary and distracting to be constantly told
  that e.g. a certain variable is a constant DWORD.  One stylistic
  convention we try to use religiously may need pointing out: we
  *always* give function arguments names starting with "a".  We find
  it extremely useful to know at a glance that a variable is or is not
  passed to the function by its caller.  (Thanks, Jim Hollenberg!)
*/

#define   DEVICE_MAIN
#include  "billgps.h"
#undef    DEVICE_MAIN

#define IPPROTO_UDP 17          /* user datagram protocol */

enum {
  NetOrderPortBootpServer = 0x4300,
  NetOrderPortBootpClient = 0x4400
};

#define Ndis_DEVICE_ID 0x28
#define SendSvcNo 45
#define NdisSendOrdinal ((Ndis_DEVICE_ID<<16) + SendSvcNo)

Declare_Virtual_Device(BILLGPS)

DefineControlHandler(SYS_DYNAMIC_DEVICE_INIT, OnSysDynamicDeviceInit);
DefineControlHandler(SYS_DYNAMIC_DEVICE_EXIT, OnSysDynamicDeviceExit);
DefineControlHandler(W32_DEVICEIOCONTROL, OnW32Deviceiocontrol);

/*
   Locked Data
*/

ProtocolBlock billgpsContext;

/*
  Note: We once tested the interaction between compiler, linker, etc. in
  getting static data right by changing NdisMedium802_3 to NdisMediumLocalTalk
  in mediumArray; it caused the havoc we expected!
*/

enum { OurMediaMax = 2 };
NDIS_MEDIUM mediumArray[OurMediaMax] = { // shouldn't be on the stack!
  NdisMedium802_3, NdisMedium802_5
};

PVOID NdisSendService;
DeviceServiceEx_THUNK thunkNdisSendHook;

/*
   Locked Code
*/

/*
  Our functions for locking and unlocking memory are slight modifications
  of Ingle/Chlap.
*/

enum { PageMask = 0xfff, PageShift = 12 };

DWORD _stdcall BgpsPageLock(DWORD amemP, DWORD asize)
{
  DWORD linOffset = amemP & PageMask; // page offset of memory to map
  DWORD linPageNum = amemP >> PageShift;  // generate page number

  // Calculate # of pages to map globally

  DWORD nPages = ((amemP + asize) >> PageShift) - linPageNum + 1;

  //
  // Return global mapping of passed in pointer, as this new pointer
  // is how the memory must be accessed out of context.
  //

  return (_LinPageLock(linPageNum, nPages, PAGEMAPGLOBAL) + linOffset);
}

void _stdcall BgpsPageUnlock(PVOID amemP, DWORD asize)
{
  DWORD linPageNum = ((UINT)amemP) >> PageShift;
  DWORD nPages = (((UINT)amemP + asize) >> PageShift) - linPageNum + 1;

  // Free globally mapped memory

  _LinPageUnLock(linPageNum, nPages, PAGEMAPGLOBAL);
}

BOOL OnSysDynamicDeviceInit()
{
  NDIS_STATUS status;

  DEBUGTRACE("OnSysDynamicDeviceInit Entry...");
  SETDEBUGLEVEL(DBG_ERROR, DBG_TRACE);

  return (DriverEntry(NULL, NULL) == NDIS_STATUS_SUCCESS);
}

BOOL OnSysDynamicDeviceExit()
{
  DEBUGTRACE("OnSysDynamicDeviceExit Entry...");

  return TRUE;
}

/*
  This function, which returns the checksum of an IP header, was adapted
  from cksum() in Comer & Stevens, Internetworking with TCP/IP, v.2,
  section 5.4.4.
*/

USHORT ipChecksum(IPHeader* anipHeaderP)
{
  USHORT ipChecksumSave = anipHeaderP->checksum; /* temporary */
  ULONG sum;
  USHORT* wordP = (USHORT*)anipHeaderP;
  UINT nWords = (anipHeaderP->versionAndHeaderLength & 0xf) * 2;
  anipHeaderP->checksum = 0;    /* Initialize embedded checksum. */
  for (sum = 0; nWords > 0; nWords--, wordP++) {
    sum += (((*wordP & 0xff) << 8) + ((*wordP & 0xff00) >> 8));
  }
  anipHeaderP->checksum = ipChecksumSave; /* temporary */
  sum = (sum >> 16) + (sum & 0xffff); /* Add in carry. */
  sum += (sum >> 16);           /* Maybe one more. */
  sum = ~sum;
  return (((sum & 0xff) << 8) + ((sum & 0xff00) >> 8));
}

/*
  Here's where we trap the outgoing bootp request and zero its IP source
  address, adjusting the UDP and IP checksums accordingly.

  The slightly tricky part of this is recognizing the bootp request as
  cooked by the Microsoft WinSock and IP stack.  Obviously, in the absence
  of inside information, this had to be done empirically and (in the
  absence of a Ring 0 debugger) iteratively.  Fortunately Microsoft
  breaks the packet into predictable chunks in which the IP header, UDP
  header and data payload are separate buffers.  This lets us base our
  test on finding the distinctive signature of the UDP header: both
  protocol ports and length.  Then we check the previous buffer to see if
  it checks out on a length basis.  Then, and only then, we proceed with
  hacking the packet.
*/

BOOL __stdcall NdisSendHook(PDSFRAME apDS)
{
  enum { SendStatusArg, SendBindingHandleArg, SendPacketArg };
  PNDIS_PACKET packetP = (PNDIS_PACKET) apDS->ARGS[SendPacketArg];
  if (packetP) {                /* We extracted a packet from the args. */
    PNDIS_BUFFER bufferP = packetP->Private.Head; /* start of buffer chain */
    PNDIS_BUFFER lastBufferP = NULL;
    UINT bufferNo = 0;
    HOOK_PROGRESS(GotPacket);
    for (; bufferP; bufferNo++, lastBufferP = bufferP,
           bufferP = (PNDIS_BUFFER)bufferP->Next) {
      UCHAR* bufferData = (UCHAR*)bufferP->VirtualAddress;
      HOOK_PROGRESS(GotBuffer);
#if 0                           /* This might work only for Ethernet */
      if (bufferNo == 0) {
        const char broadcastMACAddress[MacAddressSize] = {
          0xff, 0xff, 0xff, 0xff, 0xff, 0xff
        };
        if (!memcmp(bufferData, broadcastMACAddress, MacAddressSize)) {
          HOOK_PROGRESS(GotBroadcast);
        }
        else break;             /* Bootp request MUST be a broadcast! */ 
      }
#endif
      if (bufferNo < NHookBufferEntries) {
        billgpsContext.bufferLengths[bufferNo] = bufferP->Length;
      }
      if ((bufferP->Length == sizeof(UDPHeader)) && (lastBufferP != NULL)) {
        const char bootpRequestSignature[6] = { 0, 0x44, 0, 0x43, 1, 0x34 };
        UDPHeader* udpHeaderP = (UDPHeader*)bufferData;
        IPHeader* ipHeaderP = (IPHeader*)lastBufferP->VirtualAddress;
        if ((!memcmp(bufferData, bootpRequestSignature,
                    sizeof(bootpRequestSignature))) &&
            (lastBufferP->Length ==
             ((ipHeaderP->versionAndHeaderLength & 0xf) * 4))) {
          HOOK_PROGRESS(GotRequest);
          ipHeaderP->sourceAddr = 0; /* This is the whole point! */
          udpHeaderP->udpChecksum = 0; /* Let's make life easy! */
          ipHeaderP->checksum = ipChecksum(ipHeaderP);
        }
      }
    }
  }
  return FALSE;                 /* Please go ahead with NdisSend! */
}

/*
  Ask the adapter for its MAC address with an NdisRequest.
*/

void requestMACAddress(OpenInstance* aopenInstP)
{
  switch(aopenInstP->u1.medium)
    {
    case NdisMedium802_3:
      aopenInstP->request.DATA.QUERY_INFORMATION.Oid =
        OID_802_3_CURRENT_ADDRESS;
      break;
    case NdisMedium802_5:
      aopenInstP->request.DATA.QUERY_INFORMATION.Oid =
        OID_802_5_CURRENT_ADDRESS;
      break;
    default:
      return;
    }
  aopenInstP->request.RequestType = NdisRequestQueryInformation;
  aopenInstP->request.DATA.QUERY_INFORMATION.InformationBuffer =
    aopenInstP->macAddress;
  aopenInstP->request.DATA.QUERY_INFORMATION.InformationBufferLength =
    MacAddressSize;
  NdisRequest(&aopenInstP->macAddressReqStatus,
              aopenInstP->adapterHandle, &aopenInstP->request);
}

/*
  We need to set the filter on this bound adapter.
  While we're at it, if the previous attempt to get the MAC address failed,
  we repeat it here.
*/

void SetFilter(OpenInstance* aopenP, UINT afilter)
{
  aopenP->packetFilter = afilter;
  aopenP->filterBlock.RequestType = NdisRequestSetInformation;
  aopenP->filterBlock.DATA.SET_INFORMATION.Oid = OID_GEN_CURRENT_PACKET_FILTER;
  aopenP->filterBlock.DATA.SET_INFORMATION.InformationBuffer =
    &aopenP->packetFilter;
  aopenP->filterBlock.DATA.SET_INFORMATION.InformationBufferLength =
    sizeof(aopenP->packetFilter);
  if (aopenP->adapterHandle) {
    NdisRequest(&aopenP->setFilterStatus, aopenP->adapterHandle,
                &aopenP->filterBlock);
    if (aopenP->macAddressReqStatus != NDIS_STATUS_SUCCESS) {
      requestMACAddress(aopenP);
    }
  }
}

/*
  Here we return the instance corresponding to the handle or, failing that,
  NULL.
*/

OpenInstance* getOpenInstance(PIOCTLPARAMS p)
{
  PLIST_ENTRY  pHead = &billgpsContext.OpenList;
  PLIST_ENTRY pTemp;
  POpenInstance Open;
    
  // search the list for the Open Instance containing the specified handle
            
  for (pTemp=pHead->Flink; pTemp != pHead; pTemp=pTemp->Flink) {
    Open = CONTAINING_RECORD(pTemp, OpenInstance, ListElement);        
    if (Open) {
      if((Open->tagProcess == p->dioc_ppdb) &&
         (Open->hDevice == p->dioc_hDevice)) return Open;
    }
  }

  DEBUGTRACE("GetOpen: returning NULL\n");
  return (POpenInstance) NULL;               // just in case
}

/*
   We turn off the packet spigot by zeroing the filter.
*/

void BgpsDisconnect(PIOCTLPARAMS aPIP)
{
  OpenInstance* openInstanceP = getOpenInstance(aPIP);
  if (openInstanceP) {
    SetFilter(openInstanceP, 0);
    BgpsUnbindAdapterHandler(&openInstanceP->unbindStatus,
                             openInstanceP, NULL);
  }
}

/*
  To find the MAC address we're interested in on behalf of the app, we may
  need to scan more than one binding context before we find one that's for
  a real LAN card.
*/

DWORD getMacAddress(PIOCTLPARAMS aPIP)
{
  OpenInstance* openInstanceP = getOpenInstance(aPIP);

  if ((aPIP->dioc_OutBuf == NULL) || (aPIP->dioc_cbOutBuf < MacAddressSize) ||
      (openInstanceP == NULL)) {
    return (DEVIOCTL_NOERROR - 1);
  }
  else {
    memcpy(aPIP->dioc_OutBuf, openInstanceP->macAddress, MacAddressSize);
    *aPIP->dioc_bytesret = MacAddressSize;
  }
  return DEVIOCTL_NOERROR;
}

/*
  This is the essential part of the dynamic version of our VxD.  Here we
  do what NDIS 3.1 normally does when Windows gets loaded: we call the
  VxD's bind handler with the name of the adapter.  We rely on the app
  to get the name right, which is normally not what one would want to
  impose on application code but, hey, our app is billgPC!
*/

DWORD bindMac(PIOCTLPARAMS aPIP)
{
  if ((aPIP->dioc_InBuf == NULL) || (aPIP->dioc_cbInBuf < 1) ||
      (aPIP->dioc_cbInBuf > AdapterNameMax) ||
      (aPIP->dioc_ovrlp == NULL)) {
    return (DEVIOCTL_NOERROR - 1);
  }
  else {
    NDIS_STRING tempAdapterName;
    OpenInstance* openInstanceP;
    tempAdapterName.Length = aPIP->dioc_cbInBuf;
    NdisAllocateMemory((PVOID *)&openInstanceP, sizeof(OpenInstance), 0, -1);
    if (openInstanceP == NULL) {
      // not enough memory
      DEBUGTRACE("bindMac: ERROR: no memory for Open\n"); 
      return (DEVIOCTL_NOERROR - 1);
    }
    NdisZeroMemory((PVOID)openInstanceP, sizeof(OpenInstance));
    openInstanceP->bindEvent = aPIP->dioc_ovrlp->O_Internal; // for completion
    openInstanceP->tagProcess = aPIP->dioc_ppdb;
    openInstanceP->hDevice = aPIP->dioc_hDevice;
    // **** Conpack never calls GetOverlappedResult unless ioctl fails,
    // which it never does on bind; vpacket bind doesn't bother w/ovrlp.
    tempAdapterName.MaximumLength = aPIP->dioc_cbInBuf;
    tempAdapterName.Buffer = aPIP->dioc_InBuf;
    // Is the protocol handle correct here? ***
    BgpsBindAdapterHandler(&openInstanceP->openAdapterCompleteStatus,
                           NULL, // Don't call CompleteBindAdapter!
                           &tempAdapterName, openInstanceP, NULL);
    if ((openInstanceP->openAdapterCompleteStatus != NDIS_STATUS_SUCCESS) &&
        (openInstanceP->openAdapterCompleteStatus != NDIS_STATUS_PENDING)) {
      return (DEVIOCTL_NOERROR - 1);
    }
  }
  return DEVIOCTL_NOERROR;
}

/*
  Here we translate the NDIS medium type into the bootp hardware type
  and plunk it into the app's single-byte buffer.  We've no need for
  input or output length here.
*/

DWORD getHWType(PIOCTLPARAMS aPIP)
{
  OpenInstance* openInstanceP = getOpenInstance(aPIP);

  if ((aPIP->dioc_OutBuf == NULL) || (openInstanceP == NULL)) {
    return (DEVIOCTL_NOERROR - 1);
  }
  else {
    switch((NDIS_MEDIUM)(openInstanceP->u1.medium)) {
    case NdisMedium802_3:
      *((UCHAR*)aPIP->dioc_OutBuf) = (UCHAR)BootpHWTypeEthernet;
      break;
    case NdisMedium802_5:
      *((UCHAR*)aPIP->dioc_OutBuf) = (UCHAR)BootpHWTypeIEEE;
      break;
    default:
      return (DEVIOCTL_NOERROR - 1);
    }
  }
  return DEVIOCTL_NOERROR;
}

/*
  Here we try to dump the driver's guts, except for the buffer itself and
  any unused adapter contexts.
*/

DWORD dumpDriver(PIOCTLPARAMS aPIP)
{
  OpenInstance* openInstanceP = getOpenInstance(aPIP);
  if (aPIP->dioc_OutBuf == NULL) return (DEVIOCTL_NOERROR - 1);
  else {
    UINT sizeWanted = sizeof(billgpsContext) +
      (openInstanceP ?  sizeof(OpenInstance): 0);
    if (aPIP->dioc_cbOutBuf < sizeWanted) return (DEVIOCTL_NOERROR - 1);
    else *aPIP->dioc_bytesret = sizeWanted;
    memcpy(aPIP->dioc_OutBuf, &billgpsContext, sizeof(billgpsContext));
    if (openInstanceP) memcpy((char*)aPIP->dioc_OutBuf + sizeof(billgpsContext),
                              openInstanceP, sizeof(OpenInstance));
    return DEVIOCTL_NOERROR;
  }
}

NDIS_STATUS BgpsAllocatePacketBuffer(OpenInstance* aopenP,
                                     PNDIS_PACKET *apacketPP,
                                     PIOCTLPARAMS aPIP)
{
  NDIS_STATUS status;
  PNDIS_BUFFER ndisBufferP;
  PNDIS_PACKET packetP;
    
  //
  //  Try to get a packet from our list of free ones
  //

  NdisAllocatePacket(&status, apacketPP, aopenP->PacketPool);

  if (status != NDIS_STATUS_SUCCESS) {
    DEBUGTRACE("BgpsAllocatePacketBuffer: ERROR: no packet buffer\n");
    *(DWORD *)(aPIP->dioc_bytesret) = 0;
    return status;
  }
    
  packetP = *apacketPP;

  //
  // Buffers used asynchronously must be page locked
  //

  switch (aPIP->dioc_IOCtlCode) {
  case DIOC_BillgpAwaitBootp:
  case DIOC_BillgpAwaitAny:
    RESERVED(packetP)->lpBuffer =
      (PVOID)BgpsPageLock((DWORD) aPIP->dioc_OutBuf, aPIP->dioc_cbOutBuf);
    RESERVED(packetP)->cbBuffer = aPIP->dioc_cbOutBuf;
    break;

#if 0
  case IOCTL_PROTOCOL_WRITE:
    RESERVED(packetP)->lpBuffer =
      (PVOID)BgpsPageLock(aPIP->dioc_InBuf, aPIP->dioc_cbInBuf);
    RESERVED(packetP)->cbBuffer = aPIP->dioc_cbInBuf;
    break;
#endif

  default:
    DEBUGTRACE("BgpsAllocatePacketBuffer: ERROR: Bad FunctionCode\n");
    // recycle the packet
    NdisReinitializePacket(packetP);
            
    // Put the packet on the free queue
    NdisFreePacket(packetP);
	
    *(DWORD *)(aPIP->dioc_bytesret) = 0;
    status = NDIS_STATUS_NOT_ACCEPTED;
    return status;
  }
	
  RESERVED(packetP)->lpcbBytesReturned =
    (PVOID)BgpsPageLock((DWORD) aPIP->dioc_bytesret, sizeof(DWORD));
  RESERVED(packetP)->lpoOverlapped =
    (PVOID)BgpsPageLock((DWORD) aPIP->dioc_ovrlp, sizeof(OVERLAPPED));
  RESERVED(packetP)->hDevice = aPIP->dioc_hDevice;
  RESERVED(packetP)->tagProcess = aPIP->dioc_ppdb;

  switch (aPIP->dioc_IOCtlCode) {
  case DIOC_BillgpAwaitBootp:
  case DIOC_BillgpAwaitAny:
    NdisAllocateBuffer(&status, &ndisBufferP, aopenP->BufferPool, 
                       (PVOID)(RESERVED(packetP)->lpBuffer),
                       aPIP->dioc_cbOutBuf);
    break;

#if 0
  case IOCTL_PROTOCOL_WRITE:
    NdisAllocateBuffer(&status, &ndisBufferP, aopenP->BufferPool, 
                       (PVOID)RESERVED(packetP)->lpBuffer,
                       aPIP->cbInBuffer);
    break;
#endif
  }

  if (status != NDIS_STATUS_SUCCESS) {
    DEBUGTRACE("BgpsAllocatePacketBuffer: ERROR: no buffer descriptor\n");

    // recycle the packet

    NdisReinitializePacket(packetP);

    // Put the packet on the free queue

    NdisFreePacket(packetP);
	
    *(DWORD *)(aPIP->dioc_bytesret) = 0;
    return status;
  }

  // Attach buffer to Packet

  NdisChainBufferAtFront(packetP, ndisBufferP);

  return status;
}

DWORD BgpsRead(PIOCTLPARAMS aPIP)
{
  OpenInstance* openInstanceP = getOpenInstance(aPIP);
  PNDIS_PACKET packetP;

  if (!openInstanceP) return (DEVIOCTL_NOERROR - 1);

  if (aPIP->dioc_cbOutBuf < sizeof(BillgpBuf)) {
    *(DWORD *)(aPIP->dioc_bytesret) = 0;
    DEBUGTRACE("BgpsRead: ERROR: buffer length short\n"); 
    return (DEVIOCTL_NOERROR - 1);
  }

  if (aPIP->dioc_IOCtlCode == DIOC_BillgpAwaitAny) {
    if (!openInstanceP->promiscuous) {
      openInstanceP->promiscuous = TRUE;
      SetFilter(openInstanceP, NDIS_PACKET_TYPE_PROMISCUOUS);
    }
  }

  if (BgpsAllocatePacketBuffer(openInstanceP, &packetP, aPIP)
      == NDIS_STATUS_SUCCESS) {
    //
    //  Put this packet in a list of pending reads.
    //  The receive indication handler will attempt to remove packets
    //  from this list for use in transfer data calls
    NdisAcquireSpinLock(&openInstanceP->RcvQSpinLock);

    InsertTailList(&openInstanceP->RcvList, &RESERVED(packetP)->listElement);

    NdisReleaseSpinLock(&openInstanceP->RcvQSpinLock);
  }
  else DEBUGTRACE("BgpsRead: ERROR: allocate packet buffer failed\n");
  return DEVIOCTL_NOERROR;
}

/*

   OnW32Deviceiocontrol

   Entry point for Win32 requests. The following functions are supported:

   OPEN (CreateFile in Win32 API): connect to the app and start listening
   for packets.

   CLOSEHANDLE (CloseHandle).

   DIOC_BillgpAwaitBootp: App is ready for a bootp packet.

   DIOC_BillgpAwaitAny: App is ready for a packet, any packet!

   DIOC_BillgpGetVersion: What version of the driver is this?

   DIOC_BillgpGetMacAddress: MAC address of first (only?) netcard, please!

   DIOC_BillgpGetHWType: bootp (not NDIS) hardware type of MAC.

   DIOC_BillgpDumpDriver: Blow the driver's data into the app's buffer.
*/

DWORD OnW32Deviceiocontrol(PIOCTLPARAMS aPIP)
{
  /* Sanity Check */
  Assert( aPIP );

  if(!aPIP) return (DEVIOCTL_NOERROR - 1);

#ifdef DEBUG
  dprintf("BILLGPS: entered with function=%d\n", aPIP->dioc_IOCtlCode);
#endif

  switch (aPIP->dioc_IOCtlCode) {

  case DIOC_CLOSEHANDLE:
    BgpsDisconnect(aPIP);
    return DEVIOCTL_NOERROR;

  case DIOC_OPEN:
    return DEVIOCTL_NOERROR;

  case DIOC_BillgpAwaitBootp:
  case DIOC_BillgpAwaitAny:
    return BgpsRead(aPIP);

  case DIOC_BillgpGetVersion:
    if ((aPIP->dioc_OutBuf == NULL) || (aPIP->dioc_cbOutBuf < sizeof(UINT))) {
      return (DEVIOCTL_NOERROR - 1);
    }
    *((UINT*)aPIP->dioc_OutBuf) = BILLGPS_Major * 100 + BILLGPS_Minor;
    *aPIP->dioc_bytesret = sizeof(UINT);
    return DEVIOCTL_NOERROR;

  case DIOC_BillgpGetMacAddress:
    return getMacAddress(aPIP);

  case DIOC_BillgpGetHWType:
    return getHWType(aPIP);

  case DIOC_BillgpDumpDriver:
    return dumpDriver(aPIP);

  case DIOC_BillgpBind:
    return bindMac(aPIP);

  default:
    return (DEVIOCTL_NOERROR - 1);
  }

}

BOOL __cdecl ControlDispatcher(DWORD dwControlMessage,
                               DWORD EBX,
                               DWORD EDX,
                               DWORD ESI,
                               DWORD EDI,
                               DWORD ECX)
{
  START_CONTROL_DISPATCH

    ON_SYS_DYNAMIC_DEVICE_INIT(OnSysDynamicDeviceInit);
    ON_SYS_DYNAMIC_DEVICE_EXIT(OnSysDynamicDeviceExit);
    ON_W32_DEVICEIOCONTROL(OnW32Deviceiocontrol);

  END_CONTROL_DISPATCH

  return TRUE;
}

/*
   Main entry point for NDIS driver; called during (dynamic) initialization
*/

NTSTATUS NDIS_API DriverEntry(IN PDRIVER_OBJECT aDriverObject, // NULL for VxDs
                              IN PUNICODE_STRING aRegistryPath) // ditto
{
  NDIS_STRING BgpsName = NDIS_STRING_CONST(TRANSPORT_NAME);
  static UINT initCount = 0;

  DEBUGTRACE("DriverEntry Entry...");

  /* We don't want to register our protocol etc. twice! */
  if (initCount++) return NDIS_STATUS_SUCCESS;

  /*
     Under NDIS 3.1 PROTOCOL drivers are dynamically loaded and start
     executing by receiving the Sys_Dynamic_Init message. They then
     may call NdisOpenProtocolConfiguration() and NdisReadConfiguration()
     to to get global configuration information from the Registry.
  */

  /*
    Initialize the PROTOCOL characteristics for the call to
    NdisRegisterProtocol.
  */

  NdisZeroMemory(&billgpsContext, sizeof(billgpsContext));
  billgpsContext.protocolCharacteristics.MajorNdisVersion =
    BILLGPS_NDIS_MAJOR_VERSION;
  billgpsContext.protocolCharacteristics.MinorNdisVersion =
    BILLGPS_NDIS_MINOR_VERSION;
  billgpsContext.protocolCharacteristics.Reserved = 0;
  billgpsContext.protocolCharacteristics.OpenAdapterCompleteHandler =
    BgpsOpenAdapterCompleteHandler;
  billgpsContext.protocolCharacteristics.CloseAdapterCompleteHandler =
    BgpsCloseAdapterCompleteHandler;
  billgpsContext.protocolCharacteristics.SendCompleteHandler =
    BgpsSendCompleteHandler;
  billgpsContext.protocolCharacteristics.TransferDataCompleteHandler =
    BgpsTransferDataCompleteHandler;
  billgpsContext.protocolCharacteristics.ResetCompleteHandler =
    BgpsResetCompleteHandler;
  billgpsContext.protocolCharacteristics.RequestCompleteHandler =
    BgpsRequestCompleteHandler;
  billgpsContext.protocolCharacteristics.ReceiveHandler =
    BgpsReceiveHandler;
  billgpsContext.protocolCharacteristics.ReceiveCompleteHandler =
    BgpsReceiveCompleteHandler;
  billgpsContext.protocolCharacteristics.StatusHandler = BgpsStatusHandler;
  billgpsContext.protocolCharacteristics.StatusCompleteHandler =
    BgpsStatusCompleteHandler;
  billgpsContext.protocolCharacteristics.BindAdapterHandler =
    BgpsBindAdapterHandler;
  billgpsContext.protocolCharacteristics.UnbindAdapterHandler =
    BgpsUnbindAdapterHandler;
  billgpsContext.protocolCharacteristics.UnloadProtocolHandler =
    BgpsUnloadProtocolHandler;
  billgpsContext.protocolCharacteristics.Name = BgpsName;

  /*
     The NdisRegisterProtocol request is called when a PROTOCOL module
     initializes. It provides the NDIS interface with information about the
     PROTOCOL driver, including the addresses of its request handlers.

     The PROTOCOL driver passes in a characteristics table. This table is
     copied by the NdisRegisterProtocol request to its own internal storage.
     Thus, once registered, the PROTOCOL driver cannot alter its handler
     routines.
  */

  DEBUGTRACE("Calling NdisRegisterProtocol...");

  NdisRegisterProtocol(&billgpsContext.registerStatus,
                       &billgpsContext.ndisProtocolHandle,
                       &billgpsContext.protocolCharacteristics,
                       sizeof(NDIS_PROTOCOL_CHARACTERISTICS));

  if(billgpsContext.registerStatus != NDIS_STATUS_SUCCESS) {
    // NdisRegisterProtocol failed.
#ifdef DEBUG
    dprintf("NdisRegisterProtocol failed with code 0x%X\n",
            billgpsContext.registerStatus);
#endif
  }
  else {
    DEBUGTRACE("NdisRegisterProtocol succeeded\n");
    InitializeListHead(&billgpsContext.OpenList);
    // Hook NdisSend.
    NdisSendService = Hook_Device_Service_Ex(NdisSendOrdinal, NdisSendHook,
                                             &thunkNdisSendHook);
  }

  return billgpsContext.registerStatus;
}

/*
  If the original open was asynchronous we'll need to call
  NdisCompleteBindAdapter.
  After successfully opening the adapter, if it's a real LAN adapter we
  want to get its MAC address so the app can ask us for it later.
  We also need to set its filter so we'll see incoming packets.
*/

VOID NDIS_API
BgpsOpenAdapterCompleteHandler(IN NDIS_HANDLE abindingContext,
                               IN NDIS_STATUS astatus,
                               IN NDIS_STATUS aopenErrorStatus)
{
  OpenInstance* openInstanceP = (OpenInstance*)abindingContext;
  DEBUGTRACE("OpenAdapterCompleteHandler Entry...");
  if (openInstanceP->bindAdapterContext) {
    NdisCompleteBindAdapter(openInstanceP->bindAdapterContext,
                            astatus, aopenErrorStatus);
  }
  openInstanceP->openAdapterCompleteStatus = astatus;
  openInstanceP->openErrorStatus = aopenErrorStatus;
  if (astatus != NDIS_STATUS_SUCCESS) { // Thanks, Dan Lanciani!
    NdisFreeSpinLock(&openInstanceP->RcvQSpinLock);

    NdisFreeBufferPool(openInstanceP->BufferPool);
    NdisFreePacketPool(openInstanceP->PacketPool);

    VWIN32_DIOCCompletionRoutine(openInstanceP->bindEvent); // Tell the app.
    NdisFreeMemory(openInstanceP, sizeof(OpenInstance), 0);
    return;                     // ****Bug: won't notify of error
  }
  openInstanceP->u1.medium =
         mediumArray[openInstanceP->u1.selectedMediumIndex];
  // Insert new instance into list.
  InsertTailList(&billgpsContext.OpenList, &openInstanceP->ListElement);
  VWIN32_DIOCCompletionRoutine(openInstanceP->bindEvent);  // Tell the app.
  SetFilter(openInstanceP,
            NDIS_PACKET_TYPE_DIRECTED | NDIS_PACKET_TYPE_BROADCAST);
  requestMACAddress(openInstanceP);
}

/*
  When NDIS tells us the unbinding of an adapter's complete, we want to
  make sure we don't keep a dangling handle for it.
*/

VOID NDIS_API
BgpsCloseAdapterCompleteHandler(IN NDIS_HANDLE abindingContext,
                                IN NDIS_STATUS astatus)
{
  OpenInstance* openInstanceP = (OpenInstance*)abindingContext;
  DEBUGTRACE("CloseAdapterCompleteHandler Entry...");
  if ((astatus == NDIS_STATUS_PENDING) &&
      (openInstanceP->bindAdapterContext)) {
    // This won't happen in the dynamic version of the protocol.
    NdisCompleteUnbindAdapter(openInstanceP->bindAdapterContext, astatus);
  }

  if (astatus == NDIS_STATUS_SUCCESS) {
    RemoveEntryList(&openInstanceP->ListElement); // Remove adapter from list
    // Free Memory

    NdisFreeSpinLock(&openInstanceP->RcvQSpinLock);
    NdisFreeBufferPool(openInstanceP->BufferPool);
    NdisFreePacketPool(openInstanceP->PacketPool);
    NdisFreeMemory(openInstanceP, sizeof(OpenInstance), 0);
  }
}

VOID NDIS_API
BgpsSendCompleteHandler(IN NDIS_HANDLE abindingContext,
                        IN PNDIS_PACKET apacket,
                        IN NDIS_STATUS astatus)
{
  DEBUGTRACE("SendCompleteHandler Entry...");
}

/*
   The NDIS wrapper calls this only on an asynchronous transfer, i.e., only
   if our receive handler's call to NdisTransferData() returned with a pending
   status.  In order to centralize completion handling, if our receive
   handler is lucky enough to call NdisTransferData() and not be told to wait,
   it calls this function as if it were the mighty wrapper.

   In fact, since this is where we recycle a receive buffer, we have to
   call this function every time we consume a receive buffer, even if we
   don't use NdisTransferData().  This happens in promiscuous mode.

   Since we allow a sloppy bootp daemon to send us broadcast packets, there's
   a possibility that this packet isn't really for us.  If we can tell that
   it isn't (which we can only do if the request for this adapter's MAC
   address we made at bind time succeeded) we bail out here.  Of course, this
   is germane only if we aren't in promiscuous mode.  By the way, the reason
   we do this checking here rather than in the receive handler is that it
   involves data that just might not be in the lookahead buffer.
*/

VOID NDIS_API
BgpsTransferDataCompleteHandler(IN NDIS_HANDLE abindingContext,
                                IN PNDIS_PACKET apacketP,
                                IN NDIS_STATUS astatus,
                                IN UINT abytesTransferred)
{
  BOOL falseAlarm = FALSE;      // TRUE if this wasn't really for us
  OpenInstance* openInstanceP = (OpenInstance*)abindingContext;
  BgpsReserved* reservedP = (BgpsReserved*)(apacketP->ProtocolReserved);
  OVERLAPPED* overlapP = ((OVERLAPPED *) (reservedP->lpoOverlapped));
  PNDIS_BUFFER ndisBufferP;
  DEBUGTRACE("TransferDataCompleteHandler Entry...");
  if ((!openInstanceP->promiscuous) &&
      (openInstanceP->macAddressReqStatus == NDIS_STATUS_SUCCESS) &&
      (astatus != NDIS_STATUS_CLOSED)) {
    BYTE* packetMACAddress = ((BootpBuf*)reservedP->lpBuffer)->chaddr;
    if (memcmp(packetMACAddress, openInstanceP->macAddress,
               MacAddressSize)) {
      InsertHeadList(&openInstanceP->RcvList, &reservedP->listElement);
      return;
    }
  }
  NdisUnchainBufferAtFront(apacketP, &ndisBufferP); // Free buffer descriptor.
  if (ndisBufferP) NdisFreeBuffer(ndisBufferP);
  *(reservedP->lpcbBytesReturned) = abytesTransferred; // Set bytes transferred.
  overlapP->O_InternalHigh = *(reservedP->lpcbBytesReturned);
  VWIN32_DIOCCompletionRoutine(overlapP->O_Internal); // signal the app

  // Unlock buffers   

  BgpsPageUnlock(reservedP->lpBuffer, reservedP->cbBuffer);
  BgpsPageUnlock(reservedP->lpcbBytesReturned, sizeof(DWORD));
  BgpsPageUnlock(reservedP->lpoOverlapped, sizeof(OVERLAPPED));

  NdisReinitializePacket(apacketP); // recycle the packet

  NdisFreePacket(apacketP);      // Put the packet on the free queue
}

VOID NDIS_API
BgpsResetCompleteHandler(IN NDIS_HANDLE abindingContext,
                         IN NDIS_STATUS astatus)
{
  DEBUGTRACE("ResetCompleteHandler Entry...");
}

VOID NDIS_API
BgpsRequestCompleteHandler(IN NDIS_HANDLE abindingContext,
                           IN PNDIS_REQUEST andisRequest,
                           IN NDIS_STATUS astatus)
{
  DEBUGTRACE("RequestCompleteHandler Entry...");
}

PNDIS_PACKET _stdcall getRcvPacket(OpenInstance* aopenP)
{
  PNDIS_PACKET packetP = NULL;
  PLIST_ENTRY listEntryP;
  PBgpsReserved reservedP;
    
  NdisAcquireSpinLock(&aopenP->RcvQSpinLock);

  if (IsListEmpty(&aopenP->RcvList)) { 
    NdisReleaseSpinLock(&aopenP->RcvQSpinLock);
    DEBUGTRACE("getRcvPacket: ERROR: RcvList empty\n");
  }
  else {
    listEntryP = RemoveHeadList(&aopenP->RcvList);
    NdisReleaseSpinLock(&aopenP->RcvQSpinLock); // ***
    reservedP = CONTAINING_RECORD(listEntryP, BgpsReserved, listElement);
    packetP  =  CONTAINING_RECORD(reservedP, NDIS_PACKET, ProtocolReserved);
  }
  
  return packetP;
}

/*
  Here we handle the packet we may be waiting for.

  First we check to see if the app's really waiting, i.e. if there's an
  event.  Then we make sure it's a recognizable bootp response.  If so,
  we move the data by NdisTransferData (this is necessary because the
  packet may be longer than the lookahead buffer.)  The processing is
  finished by the TransferDataComplete handler, which we call ourselves
  if the data transfer was finished by the time we got control back from
  NdisTransferData.
*/

const char llcSignature[3] = { 0xaa, 0xaa, 3 };

NDIS_STATUS NDIS_API
BgpsReceiveHandler(IN NDIS_HANDLE abindingContext,
                   IN NDIS_HANDLE amacReceiveContext,
                   IN PVOID aheaderBuffer,
                   IN UINT aheaderBufferSize,
                   IN PVOID alookaheadBuffer,
                   IN UINT alookaheadBufferSize,
                   IN UINT apacketSize)
{
  USHORT packetType = 0xffff;   // initialized to non-IP value
  UINT ipOffset = 0;            // LLC/SNAP will augment this.
  OpenInstance* openInstanceP = (OpenInstance*)abindingContext;
  DEBUGTRACE("ReceiveHandler Entry...");

  if (!memcmp(alookaheadBuffer, llcSignature, sizeof(llcSignature))) {
    packetType = ((LLCSNAP*)alookaheadBuffer)->packetType;
    ipOffset = sizeof(LLCSNAP);
  }
  else if (aheaderBufferSize == sizeof(RFC894Header)) {
    packetType = ((RFC894Header*)aheaderBuffer)->packetType;
  }
  else if (!openInstanceP->promiscuous) return NDIS_STATUS_SUCCESS;

  if (openInstanceP->promiscuous) {
    /*
      Poor man's packet sniffer: If packet meets our criteria, dump header
      length, header and packet "data".  If it's token ring we only compare
      on the last 5 bytes of the MAC address because of TR's habit of
      setting the most significant bit.
    */
    UINT dataRoom = sizeof(BillgpBuf) - aheaderBufferSize;
    UINT dataSize =
      (alookaheadBufferSize <= dataRoom) ? alookaheadBufferSize : dataRoom;
    UINT sourceOffset = (openInstanceP->u1.medium == NdisMedium802_3) ?
      MacAddressSize : MacAddressSize + 3;
    BYTE* ourAddress =  (openInstanceP->u1.medium == NdisMedium802_3) ?
      openInstanceP->macAddress : openInstanceP->macAddress + 1;
    size_t comparisonLen = (openInstanceP->u1.medium == NdisMedium802_3) ?
      MacAddressSize : MacAddressSize - 1;
    if ((apacketSize < 328) ||   // too small for bootp
        (apacketSize > 360) ||   // almost certainly too big
        (!memcmp(((BYTE*)aheaderBuffer) + sourceOffset, // our query!
                 ourAddress, comparisonLen))) {
      return NDIS_STATUS_SUCCESS;
    }
    else {
      PNDIS_PACKET packetP = getRcvPacket(openInstanceP);
      if (packetP == NULL) return NDIS_STATUS_SUCCESS;
      *((UINT*)RESERVED(packetP)->lpBuffer) = aheaderBufferSize;
      memcpy(RESERVED(packetP)->lpBuffer + sizeof(UINT), aheaderBuffer,
             aheaderBufferSize);
      memcpy(RESERVED(packetP)->lpBuffer + aheaderBufferSize + sizeof(UINT),
             alookaheadBuffer, dataSize);
      openInstanceP->transferDataStatus = NDIS_STATUS_SUCCESS;
      BgpsTransferDataCompleteHandler(abindingContext,
                                      packetP,
                                      openInstanceP->transferDataStatus,
                                      aheaderBufferSize + dataSize +
                                      sizeof(UINT));
    }
  }
  else if (packetType == PacketTypeIP) {
    IPHeader* ipP = (IPHeader*)((BYTE*)alookaheadBuffer + ipOffset);

    if (ipP->protocol == IPPROTO_UDP) {
      UINT ipHeaderLength = (ipP->versionAndHeaderLength & 0xf) * 4;
      UINT totalHeaderLength = ipOffset + ipHeaderLength + sizeof(UDPHeader);
      UDPHeader* udpP = (UDPHeader*)(((BYTE*)ipP)+ipHeaderLength);

      if ((totalHeaderLength < alookaheadBufferSize) &&
          (totalHeaderLength < apacketSize) &&
          (udpP->destinationPort == NetOrderPortBootpClient)) {
        UINT dataSize = apacketSize - totalHeaderLength;
        UINT bytesTransferred = 0;
        PNDIS_PACKET packetP = getRcvPacket(openInstanceP);
        if (packetP == NULL) return NDIS_STATUS_SUCCESS;

        if (dataSize > sizeof(BillgpBuf)) dataSize = sizeof(BillgpBuf);
        NdisTransferData(&openInstanceP->transferDataStatus,
                         openInstanceP->adapterHandle,
                         amacReceiveContext, totalHeaderLength, dataSize,
                         packetP,
                         &bytesTransferred);
        if (openInstanceP->transferDataStatus != NDIS_STATUS_PENDING) {
          BgpsTransferDataCompleteHandler(abindingContext,
                                          packetP,
                                          openInstanceP->transferDataStatus,
                                          bytesTransferred);
        }
      }
    }
  }

  return(NDIS_STATUS_SUCCESS);
}

VOID NDIS_API
BgpsReceiveCompleteHandler(IN NDIS_HANDLE abindingContext)
{
  DEBUGTRACE("ReceiveCompleteHandler Entry...");
}

VOID NDIS_API
BgpsStatusHandler(IN NDIS_HANDLE abindingContext,
                  IN NDIS_STATUS ageneralStatus,
                  IN PVOID astatusBuffer,
                  IN UINT astatusBufferSize)
{
  DEBUGTRACE("StatusHandler Entry...");
}

VOID NDIS_API
BgpsStatusCompleteHandler(IN NDIS_HANDLE abindingContext)
{
  DEBUGTRACE("StatusCompleteHandler Entry...");
}


/*
   BIND_ADAPTER_HANDLER, UNBIND_ADAPTER_HANDLER, UNLOAD_PROTOCOL
   are valid for NDIS 3.1 protocols
*/

VOID NDIS_API
BgpsBindAdapterHandler(OUT PNDIS_STATUS astatusP,
                       IN NDIS_HANDLE abindAdapterContext,
                       IN PNDIS_STRING anadapterName,
                       IN PVOID asystemSpecific1,
                       IN PVOID asystemSpecific2)
{
  NDIS_STATUS allocStatus;
  UINT i;
  OpenInstance* openInstanceP = (OpenInstance*)asystemSpecific1;

  DEBUGTRACE("BindAdapterHandler Entry...");

  *astatusP = NDIS_STATUS_FAILURE; // Assume Failure

  //
  // Allocate a packet pool for our xmit and receive packets
  //
  NdisAllocatePacketPool(&allocStatus,
                         &(openInstanceP->PacketPool),
                         TRANSMIT_PACKETS,
                         sizeof(BgpsReserved));

  if (allocStatus != NDIS_STATUS_SUCCESS) {

    // not enough memory
    NdisFreeMemory(openInstanceP, sizeof(OpenInstance),  0);
    *astatusP = NDIS_STATUS_RESOURCES;
    DEBUGTRACE("PacketBindAdapter: ERROR: no memory for Packet Pool\n"); 
    return;
  }

  //
  // Allocate a buffer pool for our xmit and receive buffers
  //
  NdisAllocateBufferPool(&allocStatus, &(openInstanceP->BufferPool),
                         TRANSMIT_PACKETS);

  if (allocStatus != NDIS_STATUS_SUCCESS) {

    // not enough memory

    //BUG: free packet pool first

    NdisFreeMemory(openInstanceP, sizeof(OpenInstance), 0);
    *astatusP = NDIS_STATUS_RESOURCES;
    DEBUGTRACE("PacketBindAdapter: ERROR: no memory for Buffer Pool\n"); 
    return;
  }

  //
  //  Initialize list for holding pending read requests
  //
  NdisAllocateSpinLock(&(openInstanceP->RcvQSpinLock));
  InitializeListHead(&(openInstanceP->RcvList));

  /*
     Try to open the MAC

     To do: MS Help says when we get NDIS_STATUS_PENDING we should call
     NdisCompleteBindAdapter.
  */

  openInstanceP->openErrorStatus = NDIS_STATUS_FAILURE;

  {
    UINT maxLen =
      (anadapterName->Length > (sizeof(openInstanceP->adapterName) - 1)) ?
      (sizeof(openInstanceP->adapterName) - 1) :
      anadapterName->Length;
    memcpy(openInstanceP->adapterName, anadapterName->Buffer, maxLen);
    openInstanceP->adapterName[maxLen] = 0;
  }

  /*
    One way or another, the handler for completing the opening of
    the adapter will be called; we need to supply it with a
    bind adapter context handle so it can call NdisCompleteBindAdapter
    if the netcard driver calls the completion handler.  If the handle is
    0 the handler won't call NdisCompleteBindAdapter.
  */

  openInstanceP->bindAdapterContext = abindAdapterContext;

  DEBUGTRACE("Calling NdisOpenAdapter...");

  NdisOpenAdapter(astatusP, &openInstanceP->openErrorStatus,
                  &openInstanceP->adapterHandle,
                  &openInstanceP->u1.selectedMediumIndex,
                  mediumArray, OurMediaMax,
                  billgpsContext.ndisProtocolHandle,
                  (NDIS_HANDLE)openInstanceP, // ProtocolBindingContext
                  anadapterName,  // valid only until return: Network helpfile
                  0, NULL);
  if (*astatusP != NDIS_STATUS_PENDING) {
    openInstanceP->bindAdapterContext = 0;
    BgpsOpenAdapterCompleteHandler((NDIS_HANDLE)openInstanceP,
                                   *astatusP,
                                   openInstanceP->openErrorStatus);
  }

  DEBUGTRACE("NdisOpenAdapter succeeded\n");
}


/* Note
   -------
   * It appears that UnloadProtocolHandler() is not called if this function
   * returns a value other than NDIS_STATUS_SUCCESS.
   */
VOID NDIS_API
BgpsUnbindAdapterHandler(OUT PNDIS_STATUS astatusP,
                         IN NDIS_HANDLE aprotocolBindingContext,
                         IN NDIS_HANDLE anunbindAdapterContext)
{
  OpenInstance* openInstanceP = (OpenInstance*)aprotocolBindingContext;
  PLIST_ENTRY listEntry;
  PNDIS_PACKET packetP;
  DEBUGTRACE("UnbindAdapterHandler Entry...");

  // We need to complete all pending I/O requests.
  NdisAcquireSpinLock(&openInstanceP->RcvQSpinLock);

  while (!IsListEmpty(&openInstanceP->RcvList)) {
    listEntry = RemoveHeadList(&openInstanceP->RcvList);
    NdisReleaseSpinLock(&openInstanceP->RcvQSpinLock);
    packetP = CONTAINING_RECORD(listEntry, NDIS_PACKET, ProtocolReserved);
    BgpsTransferDataCompleteHandler(openInstanceP, // Complete normally
                                    packetP,
                                    NDIS_STATUS_CLOSED, 0);
    NdisAcquireSpinLock(&openInstanceP->RcvQSpinLock);
  }

  NdisReleaseSpinLock(&openInstanceP->RcvQSpinLock);

  NdisCloseAdapter(astatusP, openInstanceP->adapterHandle);

  if(*astatusP != NDIS_STATUS_PENDING) {
    BgpsCloseAdapterCompleteHandler(aprotocolBindingContext, *astatusP);
  }
}

VOID NDIS_API BgpsUnloadProtocolHandler(void)
{
  NDIS_STATUS Status = NDIS_STATUS_FAILURE;

  Unhook_Device_Service_Ex(NdisSendOrdinal, NdisSendHook, &thunkNdisSendHook);

  NdisDeregisterProtocol(&Status, billgpsContext.ndisProtocolHandle);
}

//#if 0                           // We shouldn't need this; we're dynamic

/*
   Remaining code in this module is intialization code
*/

#include INIT_CODE_SEGMENT

/*
 Handler for Control Message DEVICE_INIT
*/

BOOL OnDeviceInit(VMHANDLE ahVM, PCHAR acommandTail)
{
  SETDEBUGLEVEL(DBG_ERROR, DBG_TRACE);

  DEBUGTRACE("OnDeviceInit Entry...");

  // For NDIS 3.1 DON'T DO ANYTHING HERE!!!

  return (DriverEntry(NULL, NULL) == NDIS_STATUS_SUCCESS);
}

//#endif
